"""
Holds Analyzer settings
"""
from dataclasses import dataclass


@dataclass
class SRAnalyzerSettings:
    """
    Holds Analyzer settings.
    """

    name: str = "SRAnalyzerDefault"
    """Holds the name of the Analyzer."""
    show_process: bool = True
    """Holds if the Analyzer should show the process.
    Not implemented yet.
    """
    save_process: bool = True
    """Holds if the Analyzer should save the process.
    Not implemented yet.
    """



"""
The interface for all metrics.
"""

from abc import ABCMeta, abstractmethod
from typing import Any


class MetricResult:
    """Hold the result of a metric."""

    def __init__(self, metric_name: str, metric_value: Any, metric_unit: str):
        self.name = metric_name
        self.value = metric_value
        self.unit = metric_unit

        self.reference_image_name: str = ""
        self.image_name: str = ""

    def __str__(self) -> str:
        """Return the string representation of the metric.

        :returns: The string representation of the metric.
        """
        value_to_str = f"{self.value:.3f}" if isinstance(
            self.value, float) else self.value
        return f"{self.name}: {value_to_str} {self.unit} " \
            f"(ref: {self.reference_image_name}, comp: {self.image_name})"

    def to_dict(self) -> dict[str, Any]:
        """Return the dictionary representation of the metric.

        :returns: The dictionary representation of the metric.
        """
        return {
            "name": self.name,
            "value": self.value,
            "unit": self.unit,
            "referance": self.reference_image_name,
            "image": self.image_name
        }

    def register_image_names(self, reference_image_name: str, image_name: str) -> None:
        """Register the image names.

        :param reference_image_name: The name of the reference image.
        :param image_name: The name of the image.
        """
        self.reference_image_name = reference_image_name
        self.image_name = image_name


class InterfaceMetric(metaclass=ABCMeta):
    """The interface for all metrics."""

    @abstractmethod
    def calculate(self, **kwargs) -> MetricResult:
        """
        The function to calculate the metric.

        :param kwargs: The keywords needed to calculate the metric.
        Check the keywords_needed property.
        :returns: The calculated metric as MetricResult.
        """

    @property
    @abstractmethod
    def keywords_needed(self) -> dict[str, type]:
        """
        The keywords needed to calculate the metric.

        :return: The keywords needed.
            The keywords needed to calculate the metric.
            {
                "keyword_name_1": keyword_type,
                "keyword_name_2": keyword_type,
                (...)
            }
        """



"""
MSE implementation as a metric.
"""
import numpy
from skimage.metrics import mean_squared_error

from harmonicsradius.image import Image
from harmonicsradius.metrics.interface_metric import InterfaceMetric, MetricResult


class MeanSquaredError(InterfaceMetric):
    """The MSE metric."""

    @property
    def keywords_needed(self) -> dict[str, type]:
        """The keywords needed to calculate the metric.

        :return: The keywords needed.
        """
        return {"y_true": Image, "y_pred": Image}

    def calculate(self, **kwargs) -> MetricResult:
        """Calculate the MSE metric.

        :param kwargs: The keywords needed to calculate the metric.
        Check the keywords_needed property.
        :return: The MSE metric in a MetricResult object.
        """
        # Check keywords.
        if not set(self.keywords_needed).issubset(kwargs):
            raise ValueError(
                "Missing keywords needed to calculate the metric.")

        # Get the parameters.
        y_true = kwargs["y_true"]
        y_pred = kwargs["y_pred"]

        # Check the types of the parameters.
        if not isinstance(y_true, Image):
            raise TypeError("y_true must be an Image.")
        if not isinstance(y_pred, Image):
            raise TypeError("y_pred must be an Image.")

        # Check the shapes of the parameters.
        if y_true.get_shape() != y_pred.get_shape():
            raise ValueError("y_true and y_pred must have the same shape.")

        # Calculate the MSE.
        y_true_array: numpy.ndarray = y_true.get_image()
        y_pred_array: numpy.ndarray = y_pred.get_image()
        calculated_mse = mean_squared_error(y_true_array, y_pred_array)

        return MetricResult(metric_name="mse", metric_value=calculated_mse, metric_unit="px^2")



"""
The API of the metrics package.
"""

from harmonicsradius.metrics.harmonics_radius import HarmonicsRadius
from harmonicsradius.metrics.mean_squared_error import MeanSquaredError
from harmonicsradius.metrics.structural_similarity_index import StructuralSimilarityIndex
from harmonicsradius.metrics.peak_signal_to_noise_ratio import PeakSignalToNoiseRatio

__all__ = [
    "HarmonicsRadius",
    "MeanSquaredError",
    "StructuralSimilarityIndex",
    "PeakSignalToNoiseRatio",
]



"""
Holds the Image class to be used in the analisys.
"""
from numpy import ndarray
from typing import Union
from harmonicsradius.utils import read_image, save_image


class Image:
    """Holds the image objects and preproceses if needed."""

    def __init__(self, image_path: Union[str, 'Image', ndarray],
                 name: str, preprocess: callable = None) -> None:
        """Constructor of the Image class.

        :param image_path: The path of the image, an Image object or
        NumPy Ndarray.
        :param name: The name of the image.
        :param preprocess: The preprocess function to be applied to the image.
        """
        self._name = name

        # Read the image if it is Image or a path.
        if isinstance(image_path, Image):
            self._path = None
            self._original_image = image_path.get_image()
        elif isinstance(image_path, str):
            self._path = image_path
            try:
                self._original_image = read_image(image_path)
            except Exception as e:
                raise ValueError(f"Error while reading the image: {e}.")
            if self._original_image is None:
                raise ValueError("Error while reading the image.")
        elif isinstance(image_path, ndarray):
            self._path = None
            self._original_image = image_path
        else:
            raise ValueError("Image path must be a string or an Image object.")

        # Preprocess the image if needed.
        self._preprocess_function = preprocess
        if self._preprocess_function is not None:
            self._image = self._preprocess_function(self._original_image)
        else:
            self._image = self._original_image

    def get_image(self) -> ndarray:
        """
        Return the image.

        :return: The image in NumPy ndarray.
        """
        return self._image

    def get_shape(self) -> tuple[int, int, int]:
        """
        Return the shape of the image.

        :return: The shape of the image as (height, width, channels).
        """
        return self._image.shape

    def get_name(self) -> str:
        """
        Return the name of the image.

        :return: The name of the image.
        """
        return self._name

    def save_image(self, path: str) -> None:
        """
        Save the image to the path.

        :param path: The path to save the image.
        """
        save_image(self._image, path)



"""
Main application script
"""
import argparse
from harmonicsradius.sr_analyzer import SRAnalyzer
from harmonicsradius.metrics import HarmonicsRadius
from harmonicsradius.image import Image


def argument_parser() -> dict[str, str]:
    """Parse the command line arguments and return the parsed arguments.
    :return: The parsed arguments.
    """
    parser = argparse.ArgumentParser(
        description="Calculate HRI95 metric for given image pairs."
    )
    parser.add_argument(
        "--true",
        "-t",
        type=str,
        help="Path to the true image",
        required=True,
        dest="true_image",
    )
    parser.add_argument(
        "--predicted",
        "-p",
        type=str,
        help="Path to the predicted image",
        required=True,
        dest="predicted_image",
    )

    arguments = parser.parse_args()
    return {"true": arguments.true_image,
            "predicted": arguments.predicted_image}


def main() -> None:
    """Main function of the application."""
    # Return the file paths of the images.
    images = argument_parser()

    # Read the images.
    true_image = Image(images['true'], name="true_image")
    predicted_image = Image(images['predicted'], name="predicted_image")

    # Create the analyzer.
    analyzer = SRAnalyzer()

    # Add metrics.
    analyzer.add_metric(HarmonicsRadius())

    # Add images.
    analyzer.add_reference_image(true_image)
    analyzer.add_image(predicted_image)

    # Calculate the metrics.
    [hri95_result] = analyzer.calculate()
    print(f"{hri95_result.value:.2f}%")


if __name__ == "__main__":
    main()



"""
Holds the utility functions to be used in the analisys.
"""
import cv2
from numpy import ndarray, fft
from numpy import log as np_log
from numpy import abs as np_abs
from numpy import max as np_max
from numpy import uint8 as np_uint8


def read_image(image_path: str) -> ndarray:
    """Read the image from the path.

    :param image_path: The path of the image.
    :return: The image as a numpy array.
    """
    return cv2.imread(image_path)  # pylint: disable=no-member


def show_image(image: ndarray, title: str = "Image") -> None:
    """Show the image.

    :param image: The image to be shown.
    :param title: The title of the image.
    """
    cv2.imshow(title, image)    # pylint: disable=no-member
    cv2.waitKey(0)              # pylint: disable=no-member
    cv2.destroyAllWindows()     # pylint: disable=no-member


def save_image(image: ndarray, image_path: str) -> None:
    """Save the image to the path.

    :param image: The image to be saved.
    :param image_path: The path of the image.
    """
    cv2.imwrite(image_path, image)  # pylint: disable=no-member


def get_fft_of_image(image: ndarray, scale_log: bool = True) -> ndarray:
    """Get the FFT of the image.

    :param image: The image to get the FFT.
    :param scale_log: If the FFT should be scaled logarithmically.
    :return: The FFT of the image.
    """
    if len(image.shape) == 3:
        image = cv2.cvtColor(
            image, cv2.COLOR_BGR2GRAY)  # pylint: disable=no-member
    fft_image = fft.fftshift(fft.fft2(image))

    if scale_log:
        fft_image = np_log(1 + np_abs(fft_image))
    else:
        fft_image = np_abs(fft_image)
    return fft_image


def show_fft_image(fft_image: ndarray, title: str = "FFT Image") -> None:
    """Show the FFT image.

    :param fft_image: The FFT image to be shown.
    :param title: The title of the image.
    """
    # Importing here to avoid waiting time for matplotlib load.
    from matplotlib import pyplot as plt

    fft_uint8 = (fft_image / np_max(fft_image) * 255).astype(np_uint8)
    plt.imshow(fft_uint8, cmap="gray")
    plt.title(title)
    plt.axis("off")
    plt.show()


def draw_square_from_center(image: ndarray,
                            center: tuple[int, int],
                            half_size: int,
                            color: tuple[int, int, int] = (0, 0, 255)) -> None:
    """Draw a rectangle from the center point.

    :param image: The image to draw the rectangle.
    :param center: The center of the rectangle.
    :param half_size: The half size of the rectangle.
    :param color: The color of the rectangle in (red, green, blue).
    """
    # Find the coordinates of the rectangle.
    x_center, y_center = center
    x_start = x_center - half_size
    y_start = y_center - half_size
    x_end = x_center + half_size
    y_end = y_center + half_size

    # Copy the image and convert it to RGB.
    resulted_image = image.copy()
    resulted_image = (resulted_image / np_max(resulted_image)
                      * 255).astype(np_uint8)
    resulted_image_rgb = cv2.cvtColor(
        resulted_image, cv2.COLOR_GRAY2BGR)  # pylint: disable=no-member

    # Draw the rectangle.
    cv2.rectangle(resulted_image_rgb, (y_start, x_start),
                  (y_end, x_end), color, 2)  # pylint: disable=no-member
    show_image(resulted_image_rgb)



"""
This module defines a couple of most used preprocessor functions.
"""

import cv2
from numpy import ndarray


def shrink_to(image: ndarray, width: int, height: int) -> ndarray:
    """
    Shrink the image to the given width and height.

    :param image: The image to be shrinked.
    :param width: The width of the new image.
    :param height: The height of the new image.
    :return: The shrinked image.
    """
    return cv2.resize(image, (width, height),
                      interpolation=cv2.INTER_AREA)  # pylint: disable=no-member


def linear_upscale(image: ndarray, factor: int) -> ndarray:
    """
    Upscale the image using linear interpolation.

    :param image: The image to be upscaled.
    :param factor: The factor to upscale the image.
    :return: The upscaled image.
    """
    return cv2.resize(image, (0, 0), fx=factor, fy=factor,
                      interpolation=cv2.INTER_LINEAR)  # pylint: disable=no-member


def bicubic_upscale(image: ndarray, factor: int) -> ndarray:
    """
    Upscale the image using bicubic interpolation.

    :param image: The image to be upscaled.
    :param factor: The factor to upscale the image.
    :return: The upscaled image.
    """
    return cv2.resize(image, (0, 0), fx=factor, fy=factor,
                      interpolation=cv2.INTER_CUBIC)  # pylint: disable=no-member


def nearest_upscale(image: ndarray, factor: int) -> ndarray:
    """
    Upscale the image using nearest neighbour interpolation.

    :param image: The image to be upscaled.
    :param factor: The factor to upscale the image.
    :return: The upscaled image.
    """
    return cv2.resize(image, (0, 0), fx=factor, fy=factor,
                      interpolation=cv2.INTER_NEAREST)  # pylint: disable=no-member



"""
MSE implementation as a metric.
"""
import numpy
from skimage.metrics import structural_similarity
from harmonicsradius.image import Image
from harmonicsradius.metrics.interface_metric import InterfaceMetric, MetricResult


class StructuralSimilarityIndex(InterfaceMetric):
    """The SSIM metric."""

    @property
    def keywords_needed(self) -> dict[str, type]:
        """The keywords needed to calculate the metric.

        :return: The keywords needed.
        """
        return {"y_true": Image, "y_pred": Image}

    def calculate(self, **kwargs) -> MetricResult:
        """Calculate the SSIM metric.

        :param kwargs: The keywords needed to calculate the metric.
        Check the keywords_needed property.
        :return: The SSIM metric in a MetricResult object.
        """
        # Check keywords.
        if not set(self.keywords_needed).issubset(kwargs):
            raise ValueError(
                "Missing keywords needed to calculate the metric.")

        # Get the parameters.
        y_true = kwargs["y_true"]
        y_pred = kwargs["y_pred"]

        # Check the types of the parameters.
        if not isinstance(y_true, Image):
            raise TypeError("y_true must be an Image.")
        if not isinstance(y_pred, Image):
            raise TypeError("y_pred must be an Image.")

        # Check the shapes of the parameters.
        if y_true.get_shape() != y_pred.get_shape():
            raise ValueError("y_true and y_pred must have the same shape.")

        # Calculate the SSIM.
        y_true_array: numpy.ndarray = y_true.get_image()
        y_pred_array: numpy.ndarray = y_pred.get_image()

        calculated_ssim = structural_similarity(
            y_true_array,
            y_pred_array,
            data_range=y_true_array.max() - y_true_array.min(),
            multichannel=True,
            channel_axis=2,
        )

        return MetricResult(metric_name="SSIM", metric_value=calculated_ssim, metric_unit="")



"""
Harmonics Radius implementation with SSIM as a metric.
"""
import numpy
from skimage.metrics import structural_similarity

from harmonicsradius.image import Image
from harmonicsradius.metrics.interface_metric import InterfaceMetric, MetricResult
from harmonicsradius.utils import get_fft_of_image
#  from harmonicsradius.utils import draw_square_from_center


class HarmonicsRadius(InterfaceMetric):
    """The HRI95 metric."""
    THRESHOLD: float = 0.95
    METRIC_NAME: str = f"harmonics_radius_{int(THRESHOLD*100)}"
    METRIC_UNIT: str = "%"

    @property
    def keywords_needed(self) -> dict[str, type]:
        """The keywords needed to calculate the metric.

        :return: The keywords needed.
        """
        return {"y_true": Image, "y_pred": Image}

    def calculate(self, **kwargs) -> MetricResult:
        """Calculate the HRI95.

        :param kwargs: The keywords needed to calculate the metric.
        Check the keywords_needed property.

        :return: The HRI95 in a MetricResult object.
        """
        # Check keywords.
        if not set(self.keywords_needed).issubset(kwargs):
            raise ValueError(
                "Missing keywords needed to calculate the metric.")

        # Get the parameters.
        y_true = kwargs["y_true"]
        y_pred = kwargs["y_pred"]

        # Check the types of the parameters.
        if not isinstance(y_true, Image):
            raise TypeError("y_true must be an Image.")
        if not isinstance(y_pred, Image):
            raise TypeError("y_pred must be an Image.")

        # Check the shapes of the parameters.
        if y_true.get_shape() != y_pred.get_shape():
            raise ValueError("y_true and y_pred must have the same shape.")

        # Convert images to greyscale.
        fft_of_true: numpy.ndarray = get_fft_of_image(
            y_true.get_image(), scale_log=True)
        fft_of_pred: numpy.ndarray = get_fft_of_image(
            y_pred.get_image(), scale_log=True)

        # Find the center point.
        pred_x_center = fft_of_pred.shape[0] // 2
        pred_y_center = fft_of_pred.shape[1] // 2

        # Check the smallest side of the image.
        grid_size = self.get_smallest_side_size(fft_of_pred)

        while True:
            # Get the grid.
            grid_pred = fft_of_pred[
                pred_x_center - grid_size//2: pred_x_center + grid_size//2,
                pred_y_center - grid_size//2: pred_y_center + grid_size//2,
            ]
            grid_true = fft_of_true[
                pred_x_center - grid_size//2: pred_x_center + grid_size//2,
                pred_y_center - grid_size//2: pred_y_center + grid_size//2,
            ]

            # Check the SSIM of the grid.
            ssim_result = structural_similarity(
                grid_true,
                grid_pred,
                data_range=fft_of_true.max() - fft_of_true.min(),
                multichannel=False,
            )

            # draw_square_from_center(
            #     fft_of_pred,
            #     (pred_x_center,
            #      pred_y_center),
            #     grid_size//2,
            # )
            if ssim_result > self.THRESHOLD:
                # The radius is the half of the grid size.
                found_radius: int = grid_size / 2
                image_radius: int = self.get_smallest_side_size(
                    fft_of_pred) / 2
                metric_value = (found_radius / image_radius) * 100

                return MetricResult(
                    metric_name=self.METRIC_NAME,
                    metric_value=metric_value,
                    metric_unit=self.METRIC_UNIT
                )

            # Increase the grid size by 2.
            grid_size -= 2

            # A stop condition.
            if grid_size < 7:
                break

        # If the radius is not found, return the maximum radius.
        return MetricResult(
            metric_name=self.METRIC_NAME,
            metric_value=0,
            metric_unit=self.METRIC_UNIT
        )

    @staticmethod
    def get_smallest_side_size(image: numpy.ndarray) -> int:
        """Get the smallest side size of the image.

        :param image: The image.

        :return: The smallest side size of the image.
        """
        return image.shape[0] \
            if image.shape[0] < image.shape[1] \
            else image.shape[1]



"""
The main super resolution analyzer class.
"""
from harmonicsradius.metrics.interface_metric import InterfaceMetric, MetricResult
from harmonicsradius.image import Image
from harmonicsradius.settings import SRAnalyzerSettings


class SRAnalyzer:
    """The main super resolution analyzer class."""

    def __init__(self, settings: SRAnalyzerSettings = SRAnalyzerSettings()):
        """Initialize the SRAnalyzer class.

        :param settings: The settings of the analyzer.
        """
        self._settings = settings
        self._metrics: list[InterfaceMetric] = []
        self._reference: Image = None
        self._images: list[Image] = []
        self._is_done = False

    def add_metric(self, metric: InterfaceMetric) -> None:
        """Add a metric to the analyzer.

        :param metric: The metric to be added.
        """
        self._metrics.append(metric)

    def add_reference_image(self, image: Image) -> None:
        """Add a reference image to the analyzer.

        :param image: The reference image.
        """
        self._reference = image

    def add_image(self, image: Image) -> None:
        """Add an image to the analyzer.

        :param image: The image to be added.
        """
        self._images.append(image)

    def calculate(self) -> list[MetricResult]:
        """Calculate the metrics.

        :return: The calculated metrics.
        """
        # Check if the analyzer is done.
        if self._is_done:
            raise RuntimeError("The analyzer is already done.")

        if self._reference is None:
            raise RuntimeError("The reference image is not set.")

        if len(self._images) == 0:
            raise RuntimeError("There is no image to calculate.")

        # Calculate the metrics.
        results: list[MetricResult] = []
        for metric in self._metrics:
            for image in self._images:
                keyword_args = {
                    "y_true": self._reference,
                    "y_pred": image
                }
                metric_result = metric.calculate(**keyword_args)
                metric_result.register_image_names(
                    reference_image_name=self._reference.get_name(),
                    image_name=image.get_name()
                )
                results.append(metric_result)

        # Set true if finished.
        self._is_done = True

        return results
